@@ -1,7 +1,7 @@
 use std::fs;
 
 use crate::{
-    command::{CommandManager, CommandTokenIter, CommandTokenKind, CompletionSource},
+    command::{CommandManager, CommandToken, CommandTokenIter, CommandTokenKind, CompletionSource},
     editor::KeysIterator,
     editor_utils::{hash_bytes, ReadLinePoll},
     mode::{Mode, ModeContext, ModeKind, ModeOperation, ModeState},
@@ -58,6 +58,7 @@ impl ModeState for State {
             &mut ctx.editor.string_pool,
             &ctx.editor.buffered_keys,
             keys,
+            &ctx.editor.registers,
         ) {
             ReadLinePoll::Pending => {
                 keys.put_back();
@@ -93,12 +94,10 @@ impl ModeState for State {
             ReadLinePoll::Canceled => Mode::change_to(ctx, ModeKind::default()),
             ReadLinePoll::Submitted => {
                 let input = ctx.editor.read_line.input();
-                if !input.starts_with(|c: char| c.is_ascii_whitespace()) {
-                    ctx.editor.commands.add_to_history(input);
-                }
+                ctx.editor.commands.add_to_history(input);
 
                 let command = ctx.editor.string_pool.acquire_with(input);
-                let operation = CommandManager::eval_commands_then_output(
+                let operation = CommandManager::eval_and_then_output(
                     ctx.editor,
                     ctx.platform,
                     ctx.clients,
@@ -131,19 +130,35 @@ fn apply_completion(ctx: &mut ModeContext, cursor_movement: isize) {
 }
 
 fn update_autocomplete_entries(ctx: &mut ModeContext) {
+    fn find_command_token(tokens: &mut CommandTokenIter) -> Option<CommandToken> {
+        match tokens.next() {
+            Some((CommandTokenKind::Identifier, token)) => Some(token),
+            Some((CommandTokenKind::Register, _)) => match tokens.next() {
+                Some((CommandTokenKind::Equals, _)) => match tokens.next() {
+                    Some((CommandTokenKind::Identifier, token)) => Some(token),
+                    _ => None,
+                },
+                _ => None,
+            },
+            _ => None,
+        }
+    }
+
     let state = &mut ctx.editor.mode.command_state;
 
     let input = ctx.editor.read_line.input();
-    let mut tokens = CommandTokenIter(input);
+    let mut tokens = CommandTokenIter::new(input);
 
-    let command_name = match tokens.next() {
-        Some((_, token)) => token.trim_end_matches('!'),
+    let command_name = match find_command_token(&mut tokens) {
+        Some(token) => token.as_str(input).trim_end_matches('!'),
         None => {
             ctx.editor.picker.clear();
-            state.read_state =
-                ReadCommandState::NavigatingHistory(ctx.editor.commands.history_len());
             state.completion_index = input.len();
             state.completion_source = CompletionSource::Custom(&[]);
+            if input.trim().is_empty() {
+                state.read_state =
+                    ReadCommandState::NavigatingHistory(ctx.editor.commands.history_len());
+            }
             return;
         }
     };
@@ -157,7 +172,9 @@ fn update_autocomplete_entries(ctx: &mut ModeContext) {
     let mut last_token = None;
     for token in tokens {
         match token.0 {
-            CommandTokenKind::Text => arg_count += !is_flag_value as usize,
+            CommandTokenKind::Identifier
+            | CommandTokenKind::String
+            | CommandTokenKind::Register => arg_count += !is_flag_value as usize,
             CommandTokenKind::Flag => is_flag_value = false,
             CommandTokenKind::Equals => is_flag_value = true,
             CommandTokenKind::Unterminated => arg_count += 1,
@@ -170,14 +187,16 @@ fn update_autocomplete_entries(ctx: &mut ModeContext) {
             Some((CommandTokenKind::Unterminated, _)) => (),
             None => {
                 arg_count += 1;
-                last_token = Some((CommandTokenKind::Text, &input[input.len()..]));
+                last_token = Some((CommandTokenKind::String, CommandToken { from: 0, to: 0 }));
             }
             _ => arg_count += 1,
         }
     }
 
     let (completion_source, mut pattern) = match last_token {
-        Some((CommandTokenKind::Text, token)) | Some((CommandTokenKind::Unterminated, token))
+        Some((CommandTokenKind::Identifier, token))
+        | Some((CommandTokenKind::String, token))
+        | Some((CommandTokenKind::Unterminated, token))
             if !is_flag_value =>
         {
             let mut completion_source = CompletionSource::Custom(&[]);
@@ -191,7 +210,7 @@ fn update_autocomplete_entries(ctx: &mut ModeContext) {
                     }
                 }
             }
-            (completion_source, token)
+            (completion_source, token.as_str(input))
         }
         None => (CompletionSource::Commands, command_name),
         _ => {
@@ -228,7 +247,7 @@ fn update_autocomplete_entries(ctx: &mut ModeContext) {
             }
             CompletionSource::Buffers => {
                 for buffer in ctx.editor.buffers.iter() {
-                    if let Some(path) = buffer.path().to_str() {
+                    if let Some(path) = buffer.path.to_str() {
                         ctx.editor.picker.add_custom_entry(path);
                     }
                 }
