@@ -1,4 +1,4 @@
-use std::{convert::From, fmt, ops::AddAssign};
+use std::{convert::From, fmt, ops::AddAssign, str::Chars};
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum MatchResult {
@@ -52,7 +52,7 @@ impl Pattern {
     }
 
     pub fn compile(&mut self, pattern: &str) -> Result<(), PatternError> {
-        match PatternCompiler::new(&mut self.ops, pattern.as_bytes()).compile() {
+        match PatternCompiler::new(&mut self.ops, pattern).compile() {
             Ok(start_jump) => {
                 self.start_jump = start_jump;
                 Ok(())
@@ -76,95 +76,91 @@ impl Pattern {
     }
 
     pub fn matches_with_state(&self, text: &str, state: PatternState) -> MatchResult {
-        let mut bytes = text.as_bytes();
+        let mut chars = text.chars();
         let ops = &self.ops;
         let mut op_index = state.op_index;
 
         #[inline]
-        fn check_and_jump<F>(bytes: &mut &[u8], okj: Jump, erj: Jump, predicate: F) -> usize
+        fn index(text: &str, chars: &Chars) -> usize {
+            chars.as_str().as_ptr() as usize - text.as_ptr() as usize
+        }
+
+        #[inline]
+        fn check_and_jump<F>(chars: &mut Chars, okj: Jump, erj: Jump, predicate: F) -> usize
         where
-            F: Fn(u8) -> bool,
+            F: Fn(char) -> bool,
         {
-            if !bytes.is_empty() && predicate(bytes[0]) {
-                *bytes = &bytes[1..];
-                okj.0 as _
-            } else {
-                erj.0 as _
+            let previous_state = chars.clone();
+            match chars.next() {
+                Some(c) if predicate(c) => okj.0 as _,
+                _ => {
+                    *chars = previous_state;
+                    erj.0 as _
+                }
             }
         }
 
         loop {
             match ops[op_index] {
-                Op::Ok => {
-                    return MatchResult::Ok(
-                        bytes.as_ptr() as usize - text.as_bytes().as_ptr() as usize,
-                    )
-                }
+                Op::Ok => return MatchResult::Ok(index(text, &chars)),
                 Op::Error => return MatchResult::Err,
                 Op::Reset(jump) => {
-                    bytes = text.as_bytes();
+                    chars = text.chars();
                     op_index = jump.0 as _;
                 }
                 Op::Unwind(jump, len) => {
-                    bytes = unsafe {
-                        std::slice::from_raw_parts(
-                            bytes.as_ptr().offset(-(len.0 as isize)),
-                            bytes.len() + len.0 as usize,
-                        )
+                    let len = (len.0 - 1) as _;
+                    let index = index(text, &chars);
+                    chars = match text[..index].char_indices().rev().nth(len) {
+                        Some((i, _)) => text[i..].chars(),
+                        None => unreachable!(),
                     };
                     op_index = jump.0 as _;
                 }
                 Op::EndAnchor(okj, erj) => {
-                    if bytes.is_empty() {
+                    if chars.as_str().is_empty() {
                         op_index = okj.0 as _;
                         return match ops[op_index] {
-                            Op::Ok => MatchResult::Ok(
-                                bytes.as_ptr() as usize - text.as_bytes().as_ptr() as usize,
-                            ),
+                            Op::Ok => MatchResult::Ok(index(text, &chars)),
                             _ => MatchResult::Pending(PatternState { op_index }),
                         };
                     } else {
                         op_index = erj.0 as _;
                     }
                 }
-                Op::SkipOne(okj, erj) => op_index = check_and_jump(&mut bytes, okj, erj, |_| true),
+                Op::SkipOne(okj, erj) => op_index = check_and_jump(&mut chars, okj, erj, |_| true),
                 Op::SkipMany(okj, erj, len) => {
-                    let len = len.0 as usize;
-                    if bytes.len() >= len {
-                        bytes = &bytes[len..];
-                        op_index = okj.0 as _;
-                    } else {
-                        op_index = erj.0 as _;
-                    }
+                    let len = (len.0 - 1) as _;
+                    op_index = match chars.nth(len) {
+                        Some(_) => okj.0 as _,
+                        None => erj.0 as _,
+                    };
                 }
                 Op::Alphabetic(okj, erj) => {
-                    op_index = check_and_jump(&mut bytes, okj, erj, |b| b.is_ascii_alphabetic());
+                    op_index = check_and_jump(&mut chars, okj, erj, |c| c.is_ascii_alphabetic());
                 }
                 Op::Lower(okj, erj) => {
-                    op_index = check_and_jump(&mut bytes, okj, erj, |b| b.is_ascii_lowercase());
+                    op_index = check_and_jump(&mut chars, okj, erj, |c| c.is_ascii_lowercase());
                 }
                 Op::Upper(okj, erj) => {
-                    op_index = check_and_jump(&mut bytes, okj, erj, |b| b.is_ascii_uppercase());
+                    op_index = check_and_jump(&mut chars, okj, erj, |c| c.is_ascii_uppercase());
                 }
                 Op::Digit(okj, erj) => {
-                    op_index = check_and_jump(&mut bytes, okj, erj, |b| b.is_ascii_digit());
+                    op_index = check_and_jump(&mut chars, okj, erj, |c| c.is_ascii_digit());
                 }
                 Op::Alphanumeric(okj, erj) => {
-                    op_index = check_and_jump(&mut bytes, okj, erj, |b| b.is_ascii_alphanumeric());
+                    op_index = check_and_jump(&mut chars, okj, erj, |c| c.is_ascii_alphanumeric());
                 }
-                Op::Byte(okj, erj, byte) => {
-                    op_index = check_and_jump(&mut bytes, okj, erj, |b| b == byte)
+                Op::Char(okj, erj, ch) => {
+                    op_index = check_and_jump(&mut chars, okj, erj, |c| c == ch)
                 }
-                Op::Bytes3(okj, erj, bs) => {
-                    if bytes.len() >= 3
-                        && bytes[0] == bs[0]
-                        && bytes[1] == bs[1]
-                        && bytes[2] == bs[2]
-                    {
-                        op_index = okj.0 as _;
-                        bytes = &bytes[3..];
+                Op::String(okj, erj, len, bytes) => {
+                    let s = unsafe { std::str::from_utf8_unchecked(&bytes[..len as usize]) };
+                    op_index = if chars.as_str().starts_with(s) {
+                        chars = chars.as_str()[s.len()..].chars();
+                        okj.0 as _
                     } else {
-                        op_index = erj.0 as _;
+                        erj.0 as _
                     }
                 }
             };
@@ -229,6 +225,8 @@ enum JumpFrom {
     End(Jump),
 }
 
+const OP_STRING_LEN: usize = 10;
+
 #[derive(Clone)]
 enum Op {
     Ok,
@@ -243,15 +241,15 @@ enum Op {
     Upper(Jump, Jump),
     Digit(Jump, Jump),
     Alphanumeric(Jump, Jump),
-    Byte(Jump, Jump, u8),
-    Bytes3(Jump, Jump, [u8; 3]),
+    Char(Jump, Jump, char),
+    String(Jump, Jump, u8, [u8; OP_STRING_LEN]),
 }
 
 impl fmt::Debug for Op {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         const WIDTH: usize = 14;
 
-        fn p(f: &mut fmt::Formatter, name: &str, okj: &Jump, erj: &Jump) -> fmt::Result {
+        fn p(f: &mut fmt::Formatter, name: &str, okj: Jump, erj: Jump) -> fmt::Result {
             f.write_fmt(format_args!(
                 "{:width$}{} {}",
                 name,
@@ -262,24 +260,24 @@ impl fmt::Debug for Op {
         }
 
         match self {
-            Op::Ok => f.write_str("Ok"),
-            Op::Error => f.write_str("Error"),
-            Op::Reset(jump) => f.write_fmt(format_args!(
+            &Op::Ok => f.write_str("Ok"),
+            &Op::Error => f.write_str("Error"),
+            &Op::Reset(jump) => f.write_fmt(format_args!(
                 "{:width$} {}",
                 "Reset",
                 jump.0,
                 width = WIDTH - 4,
             )),
-            Op::Unwind(jump, len) => f.write_fmt(format_args!(
+            &Op::Unwind(jump, len) => f.write_fmt(format_args!(
                 "{:width$}[{}] {}",
                 "Unwind",
                 len.0,
                 jump.0,
                 width = WIDTH - 4
             )),
-            Op::EndAnchor(okj, erj) => p(f, "EndAnchor", okj, erj),
-            Op::SkipOne(okj, erj) => p(f, "SkipOne", okj, erj),
-            Op::SkipMany(okj, erj, len) => f.write_fmt(format_args!(
+            &Op::EndAnchor(okj, erj) => p(f, "EndAnchor", okj, erj),
+            &Op::SkipOne(okj, erj) => p(f, "SkipOne", okj, erj),
+            &Op::SkipMany(okj, erj, len) => f.write_fmt(format_args!(
                 "{:width$}[{}] {} {}",
                 "SkipMany",
                 len.0,
@@ -287,25 +285,23 @@ impl fmt::Debug for Op {
                 erj.0,
                 width = WIDTH - 4
             )),
-            Op::Alphabetic(okj, erj) => p(f, "Alphabetic", okj, erj),
-            Op::Lower(okj, erj) => p(f, "Lower", okj, erj),
-            Op::Upper(okj, erj) => p(f, "Upper", okj, erj),
-            Op::Digit(okj, erj) => p(f, "Digit", okj, erj),
-            Op::Alphanumeric(okj, erj) => p(f, "Alphanumeric", okj, erj),
-            Op::Byte(okj, erj, byte) => f.write_fmt(format_args!(
+            &Op::Alphabetic(okj, erj) => p(f, "Alphabetic", okj, erj),
+            &Op::Lower(okj, erj) => p(f, "Lower", okj, erj),
+            &Op::Upper(okj, erj) => p(f, "Upper", okj, erj),
+            &Op::Digit(okj, erj) => p(f, "Digit", okj, erj),
+            &Op::Alphanumeric(okj, erj) => p(f, "Alphanumeric", okj, erj),
+            &Op::Char(okj, erj, c) => f.write_fmt(format_args!(
                 "{:width$}'{}' {} {}",
-                "Byte",
-                *byte as char,
+                "Char",
+                c,
                 okj.0,
                 erj.0,
                 width = WIDTH - 4
             )),
-            Op::Bytes3(okj, erj, bytes) => f.write_fmt(format_args!(
-                "{:width$}'{}','{}','{}' {} {}",
-                "Bytes3",
-                bytes[0] as char,
-                bytes[1] as char,
-                bytes[2] as char,
+            &Op::String(okj, erj, len, bytes) => f.write_fmt(format_args!(
+                "{:width$}'{}' {} {}",
+                "String",
+                unsafe { std::str::from_utf8_unchecked(&bytes[..len as usize]) },
                 okj.0,
                 erj.0,
                 width = WIDTH - 4
@@ -315,18 +311,18 @@ impl fmt::Debug for Op {
 }
 
 struct PatternCompiler<'a> {
-    pub bytes: &'a [u8],
-    pub index: usize,
+    pub text: Chars<'a>,
+    pub current_char: char,
     pub start_jump: Jump,
     pub ops: &'a mut Vec<Op>,
 }
 
 impl<'a> PatternCompiler<'a> {
-    pub fn new(ops: &'a mut Vec<Op>, bytes: &'a [u8]) -> Self {
+    pub fn new(ops: &'a mut Vec<Op>, text: &'a str) -> Self {
         ops.clear();
         Self {
-            bytes,
-            index: 0,
+            text: text.chars(),
+            current_char: '\0',
             start_jump: Jump(2),
             ops,
         }
@@ -340,39 +336,27 @@ impl<'a> PatternCompiler<'a> {
         Ok(self.start_jump)
     }
 
-    fn peek(&self) -> Result<u8, PatternError> {
-        if self.index < self.bytes.len() {
-            Ok(self.bytes[self.index])
-        } else {
-            Err(PatternError::UnexpectedEndOfPattern)
-        }
-    }
-
-    fn current(&self) -> u8 {
-        self.bytes[self.index - 1]
-    }
-
-    fn assert_current(&self, byte: u8) -> Result<(), PatternError> {
-        if self.current() == byte {
+    fn assert_current(&self, c: char) -> Result<(), PatternError> {
+        if self.current_char == c {
             Ok(())
         } else {
-            Err(PatternError::Expected(byte as char))
+            Err(PatternError::Expected(c))
         }
     }
 
-    fn next(&mut self) -> Result<u8, PatternError> {
-        match self.peek() {
-            Ok(b) => {
-                self.index += 1;
-                Ok(b)
+    fn next(&mut self) -> Result<char, PatternError> {
+        match self.text.next() {
+            Some(c) => {
+                self.current_char = c;
+                Ok(c)
             }
-            Err(e) => Err(e),
+            None => Err(PatternError::UnexpectedEndOfPattern),
         }
     }
 
-    fn next_is_not(&mut self, byte: u8) -> Result<bool, PatternError> {
+    fn next_is(&mut self, c: char) -> Result<bool, PatternError> {
         match self.next() {
-            Ok(b) => Ok(b != byte),
+            Ok(ch) => Ok(ch == c),
             Err(e) => Err(e),
         }
     }
@@ -398,7 +382,7 @@ impl<'a> PatternCompiler<'a> {
         if let Ok(_) = self.next() {
             self.parse_stmt(JumpFrom::Beginning(reset_jump))?;
             while let Ok(_) = self.next() {
-                if self.current() == b'|' {
+                if self.current_char == '|' {
                     self.next()?;
                     self.ops.push(Op::Unwind(Jump(1), Length(0)));
                     patch_reset_jump(self, reset_jump);
@@ -413,8 +397,8 @@ impl<'a> PatternCompiler<'a> {
     }
 
     fn parse_stmt(&mut self, erj: JumpFrom) -> Result<(), PatternError> {
-        match self.current() {
-            b'{' => self.parse_repeat_stmt(erj),
+        match self.current_char {
+            '{' => self.parse_repeat_stmt(erj),
             _ => match self.parse_expr(JumpFrom::End(Jump(0)), erj) {
                 Ok(_) => Ok(()),
                 Err(e) => Err(e),
@@ -423,9 +407,9 @@ impl<'a> PatternCompiler<'a> {
     }
 
     fn parse_expr(&mut self, okj: JumpFrom, erj: JumpFrom) -> Result<Length, PatternError> {
-        let len = match self.current() {
-            b'(' => self.parse_sequence_expr(okj, erj)?,
-            b'[' => self.parse_group_expr(okj, erj)?,
+        let len = match self.current_char {
+            '(' => self.parse_sequence_expr(okj, erj)?,
+            '[' => self.parse_group_expr(okj, erj)?,
             _ => self.parse_class_expr(okj, erj)?,
         };
 
@@ -480,9 +464,9 @@ impl<'a> PatternCompiler<'a> {
         let end_jump = self.get_absolute_jump(JumpFrom::End(Jump(0)));
 
         let mut has_cancel_pattern = false;
-        while self.next_is_not(b'}')? {
-            match self.current() {
-                b'!' => {
+        while !self.next_is('}')? {
+            match self.current_char {
+                '!' => {
                     self.next()?;
                     self.parse_expr(JumpFrom::Beginning(end_jump), JumpFrom::End(Jump(0)))?;
                     has_cancel_pattern = true;
@@ -499,7 +483,7 @@ impl<'a> PatternCompiler<'a> {
 
         self.patch_unwind_jump(JumpFrom::End(Jump(0)), end_jump);
 
-        self.assert_current(b'}')?;
+        self.assert_current('}')?;
         Ok(())
     }
 
@@ -508,14 +492,12 @@ impl<'a> PatternCompiler<'a> {
         okj: JumpFrom,
         erj: JumpFrom,
     ) -> Result<Length, PatternError> {
-        let inverse = self.peek()? == b'!';
+        let previous_state = self.text.clone();
         let mut len = Length(0);
 
-        if inverse {
-            self.next()?;
-
+        if self.next()? == '!' {
             let abs_erj = self.get_absolute_jump(erj);
-            while self.next_is_not(b')')? {
+            while !self.next_is(')')? {
                 let expr_len = self.parse_expr(JumpFrom::End(Jump(2)), JumpFrom::End(Jump(0)))?;
                 self.skip(
                     (self.ops.len() + 3).into(),
@@ -529,8 +511,9 @@ impl<'a> PatternCompiler<'a> {
             self.jump_at_end(okj);
             self.patch_unwind_jump(erj, abs_erj);
         } else {
+            self.text = previous_state;
             let abs_erj = self.get_absolute_jump(erj);
-            while self.next_is_not(b')')? {
+            while !self.next_is(')')? {
                 let expr_len = self.parse_expr(JumpFrom::End(Jump(1)), JumpFrom::End(Jump(0)))?;
                 self.ops.push(Op::Unwind(abs_erj, len));
                 len += expr_len;
@@ -539,19 +522,17 @@ impl<'a> PatternCompiler<'a> {
             self.patch_unwind_jump(erj, abs_erj);
         }
 
-        self.assert_current(b')')?;
+        self.assert_current(')')?;
         Ok(len)
     }
 
     fn parse_group_expr(&mut self, okj: JumpFrom, erj: JumpFrom) -> Result<Length, PatternError> {
-        let inverse = self.peek()? == b'!';
+        let previous_state = self.text.clone();
         let mut len = None;
 
-        if inverse {
-            self.next()?;
-
+        if self.next()? == '!' {
             let abs_erj = self.get_absolute_jump(erj);
-            while self.next_is_not(b']')? {
+            while !self.next_is(']')? {
                 let expr_len = self.parse_expr(JumpFrom::End(Jump(0)), JumpFrom::End(Jump(1)))?;
                 self.ops.push(Op::Unwind(abs_erj, expr_len));
 
@@ -571,8 +552,9 @@ impl<'a> PatternCompiler<'a> {
             }
             self.patch_unwind_jump(erj, abs_erj);
         } else {
+            self.text = previous_state;
             let abs_okj = self.get_absolute_jump(okj);
-            while self.next_is_not(b']')? {
+            while !self.next_is(']')? {
                 let expr_len =
                     self.parse_expr(JumpFrom::Beginning(abs_okj), JumpFrom::End(Jump(0)))?;
 
@@ -585,7 +567,7 @@ impl<'a> PatternCompiler<'a> {
             self.patch_unwind_jump(okj, abs_okj);
         }
 
-        self.assert_current(b']')?;
+        self.assert_current(']')?;
         len.ok_or(PatternError::EmptyGroup)
     }
 
@@ -607,40 +589,40 @@ impl<'a> PatternCompiler<'a> {
             }
         };
 
-        let op = match self.current() {
-            b'%' => match self.next()? {
-                b'a' => Op::Alphabetic(okj, erj),
-                b'l' => Op::Lower(okj, erj),
-                b'u' => Op::Upper(okj, erj),
-                b'd' => Op::Digit(okj, erj),
-                b'w' => Op::Alphanumeric(okj, erj),
-                b'%' => Op::Byte(okj, erj, b'%'),
-                b'$' => Op::Byte(okj, erj, b'$'),
-                b'.' => Op::Byte(okj, erj, b'.'),
-                b'!' => Op::Byte(okj, erj, b'!'),
-                b'(' => Op::Byte(okj, erj, b'('),
-                b')' => Op::Byte(okj, erj, b')'),
-                b'[' => Op::Byte(okj, erj, b'['),
-                b']' => Op::Byte(okj, erj, b']'),
-                b'{' => Op::Byte(okj, erj, b'{'),
-                b'}' => Op::Byte(okj, erj, b'}'),
-                b'|' => Op::Byte(okj, erj, b'|'),
-                b => return Err(PatternError::InvalidEscaping(b as char)),
+        let op = match self.current_char {
+            '%' => match self.next()? {
+                'a' => Op::Alphabetic(okj, erj),
+                'l' => Op::Lower(okj, erj),
+                'u' => Op::Upper(okj, erj),
+                'd' => Op::Digit(okj, erj),
+                'w' => Op::Alphanumeric(okj, erj),
+                '%' => Op::Char(okj, erj, '%'),
+                '$' => Op::Char(okj, erj, '$'),
+                '.' => Op::Char(okj, erj, '.'),
+                '!' => Op::Char(okj, erj, '!'),
+                '(' => Op::Char(okj, erj, '('),
+                ')' => Op::Char(okj, erj, ')'),
+                '[' => Op::Char(okj, erj, '['),
+                ']' => Op::Char(okj, erj, ']'),
+                '{' => Op::Char(okj, erj, '{'),
+                '}' => Op::Char(okj, erj, '}'),
+                '|' => Op::Char(okj, erj, '|'),
+                c => return Err(PatternError::InvalidEscaping(c)),
             },
-            b'$' => {
+            '$' => {
                 self.ops.push(Op::EndAnchor(okj, erj));
                 return Ok(Length(0));
             }
-            b'.' => Op::SkipOne(okj, erj),
-            b'!' => return Err(PatternError::Unescaped('!')),
-            b'(' => return Err(PatternError::Unescaped('(')),
-            b')' => return Err(PatternError::Unescaped(')')),
-            b'[' => return Err(PatternError::Unescaped('[')),
-            b']' => return Err(PatternError::Unescaped(']')),
-            b'{' => return Err(PatternError::Unescaped('{')),
-            b'}' => return Err(PatternError::Unescaped('}')),
-            b'|' => return Err(PatternError::Unescaped('|')),
-            b => Op::Byte(okj, erj, b),
+            '.' => Op::SkipOne(okj, erj),
+            '!' => return Err(PatternError::Unescaped('!')),
+            '(' => return Err(PatternError::Unescaped('(')),
+            ')' => return Err(PatternError::Unescaped(')')),
+            '[' => return Err(PatternError::Unescaped('[')),
+            ']' => return Err(PatternError::Unescaped(']')),
+            '{' => return Err(PatternError::Unescaped('{')),
+            '}' => return Err(PatternError::Unescaped('}')),
+            '|' => return Err(PatternError::Unescaped('|')),
+            c => Op::Char(okj, erj, c),
         };
 
         self.ops.push(op);
@@ -651,9 +633,9 @@ impl<'a> PatternCompiler<'a> {
         let mut i = 0;
         while i < self.ops.len() {
             match &self.ops[i] {
-                Op::Byte(_, _, _) => {
-                    if !self.try_collapse_bytes3_at(i) {
-                        self.try_collapse_sequence3_at(i);
+                Op::Char(_, _, _) => {
+                    if !self.try_collapse_chars_at(i) {
+                        self.try_collapse_sequence_at(i);
                     }
                     i += 1;
                 }
@@ -673,12 +655,6 @@ impl<'a> PatternCompiler<'a> {
             jump.0 -= 1;
         }
 
-        if self.start_jump.0 as usize > index {
-            self.start_jump.0 -= 1;
-        } else if self.start_jump.0 as usize == index {
-            self.start_jump = jump;
-        }
-
         #[inline]
         fn fix_jump(jump: &mut Jump, index: usize, removed_jump: Jump) {
             if jump.0 as usize > index {
@@ -688,6 +664,8 @@ impl<'a> PatternCompiler<'a> {
             }
         }
 
+        fix_jump(&mut self.start_jump, index, jump);
+
         for op in self.ops.iter_mut() {
             match op {
                 Op::Ok | Op::Error => (),
@@ -700,8 +678,8 @@ impl<'a> PatternCompiler<'a> {
                 | Op::Upper(okj, erj)
                 | Op::Digit(okj, erj)
                 | Op::Alphanumeric(okj, erj)
-                | Op::Byte(okj, erj, _)
-                | Op::Bytes3(okj, erj, _) => {
+                | Op::Char(okj, erj, _)
+                | Op::String(okj, erj, _, _) => {
                     fix_jump(okj, index, jump);
                     fix_jump(erj, index, jump);
                 }
@@ -709,47 +687,52 @@ impl<'a> PatternCompiler<'a> {
         }
     }
 
-    fn try_collapse_bytes3_at(&mut self, index: usize) -> bool {
-        if index + 3 > self.ops.len() {
+    fn try_collapse_chars_at(&mut self, index: usize) -> bool {
+        let (c, mut okj, erj) = match self.ops[index] {
+            Op::Char(okj, erj, c) => (c, okj, erj),
+            _ => return false,
+        };
+        let mut bytes = [0; OP_STRING_LEN];
+        let mut len = c.encode_utf8(&mut bytes).len();
+
+        let mut op_index = index + 1;
+        while op_index < self.ops.len() {
+            let (oj, ej, c) = match self.ops[op_index] {
+                Op::Char(oj, ej, c) => (oj, ej, c),
+                _ => break,
+            };
+            if op_index + 1 != oj.0 as _ || erj.0 != ej.0 || len + c.len_utf8() > OP_STRING_LEN {
+                break;
+            }
+
+            len += c.encode_utf8(&mut bytes[len..]).len();
+            okj = oj;
+            op_index += 1;
+        }
+
+        let from = index + 1;
+        let to = op_index;
+        if from == to {
             return false;
         }
 
-        let mut final_okj = None;
-        let mut final_erj = None;
-        let mut bytes = [0 as u8; 3];
-
-        for i in 0..bytes.len() {
-            let op_index = index + i;
-            match &self.ops[op_index] {
-                Op::Byte(okj, erj, b)
-                    if okj.0 as usize == op_index + 1 && final_erj.unwrap_or(*erj).0 == erj.0 =>
-                {
-                    bytes[i] = *b;
-                    final_okj = Some(*okj);
-                    final_erj = Some(*erj);
-                }
-                _ => return false,
-            }
-        }
-
-        self.ops[index] = Op::Bytes3(final_okj.unwrap(), final_erj.unwrap(), bytes);
-        self.ops.drain((index + 1)..(index + 3));
-
-        if self.start_jump.0 as usize > index {
-            self.start_jump.0 -= 2;
-        }
+        self.ops[index] = Op::String(okj, erj, len as _, bytes);
+        self.ops.drain(from..to);
 
         #[inline]
-        fn fix_jump(jump: &mut Jump, index: usize) {
+        fn fix_jump(jump: &mut Jump, index: usize, fix: u16) {
             if jump.0 as usize > index {
-                jump.0 -= 2;
+                jump.0 -= fix;
             }
         }
 
+        let fix = (len - 1) as _;
+        fix_jump(&mut self.start_jump, index, fix);
+
         for op in self.ops.iter_mut() {
             match op {
                 Op::Ok | Op::Error => (),
-                Op::Reset(j) | Op::Unwind(j, _) => fix_jump(j, index),
+                Op::Reset(j) | Op::Unwind(j, _) => fix_jump(j, index, fix),
                 Op::EndAnchor(okj, erj)
                 | Op::SkipOne(okj, erj)
                 | Op::SkipMany(okj, erj, _)
@@ -758,10 +741,10 @@ impl<'a> PatternCompiler<'a> {
                 | Op::Upper(okj, erj)
                 | Op::Digit(okj, erj)
                 | Op::Alphanumeric(okj, erj)
-                | Op::Byte(okj, erj, _)
-                | Op::Bytes3(okj, erj, _) => {
-                    fix_jump(okj, index);
-                    fix_jump(erj, index);
+                | Op::Char(okj, erj, _)
+                | Op::String(okj, erj, _, _) => {
+                    fix_jump(okj, index, fix);
+                    fix_jump(erj, index, fix);
                 }
             }
         }
@@ -769,54 +752,66 @@ impl<'a> PatternCompiler<'a> {
         true
     }
 
-    fn try_collapse_sequence3_at(&mut self, index: usize) -> bool {
-        if index + 6 > self.ops.len() {
-            return false;
-        }
+    fn try_collapse_sequence_at(&mut self, index: usize) {
+        let mut bytes = [0; OP_STRING_LEN];
+        let mut len = 0;
 
-        let mut final_okj = None;
-        let mut final_erj = None;
-        let mut bytes = [0 as u8; 3];
+        let mut jumps = None;
 
-        for i in 0..bytes.len() {
-            let op_index = index + i * 2;
-            match &self.ops[op_index] {
-                Op::Byte(okj, erj, b)
-                    if okj.0 as usize == op_index + 2 && erj.0 as usize == op_index + 1 =>
-                {
-                    bytes[i] = *b;
-                    final_okj = Some(*okj);
-                }
-                _ => return false,
+        let mut sequence_len = 0;
+        let mut op_index = index;
+        while op_index + 1 < self.ops.len() {
+            let (okj, erj, c) = match self.ops[op_index] {
+                Op::Char(okj, erj, c) => (okj, erj, c),
+                _ => break,
+            };
+            if op_index + 2 != okj.0 as _
+                || op_index + 1 != erj.0 as _
+                || len + c.len_utf8() > OP_STRING_LEN
+            {
+                break;
             }
 
-            let op_index = op_index + 1;
-            match &self.ops[op_index] {
-                Op::Unwind(jump, len) if len.0 as usize == i => {
-                    final_erj = Some(*jump);
-                }
-                _ => return false,
+            let (jump, count) = match self.ops[op_index + 1] {
+                Op::Unwind(jump, count) => (jump, count),
+                _ => break,
+            };
+            if sequence_len != count.0 as _ {
+                break;
             }
+
+            len += c.encode_utf8(&mut bytes[len..]).len();
+            jumps = Some((okj, jump));
+            sequence_len += 1;
+            op_index += 2;
         }
 
-        self.ops[index] = Op::Bytes3(final_okj.unwrap(), final_erj.unwrap(), bytes);
-        self.ops.drain((index + 1)..(index + 6));
-
-        if self.start_jump.0 as usize > index {
-            self.start_jump.0 -= 5;
+        if sequence_len <= 1 {
+            return;
         }
 
+        let (okj, erj) = match jumps {
+            Some(jumps) => jumps,
+            None => return,
+        };
+
+        self.ops[index] = Op::String(okj, erj, len as _, bytes);
+        self.ops.drain(index + 1..op_index);
+
         #[inline]
-        fn fix_jump(jump: &mut Jump, index: usize) {
+        fn fix_jump(jump: &mut Jump, index: usize, fix: u16) {
             if jump.0 as usize > index {
-                jump.0 -= 5;
+                jump.0 -= fix;
             }
         }
 
+        let fix = (sequence_len * 2 - 1) as _;
+        fix_jump(&mut self.start_jump, index, fix);
+
         for op in self.ops.iter_mut() {
             match op {
                 Op::Ok | Op::Error => (),
-                Op::Reset(j) | Op::Unwind(j, _) => fix_jump(j, index),
+                Op::Reset(j) | Op::Unwind(j, _) => fix_jump(j, index, fix),
                 Op::EndAnchor(okj, erj)
                 | Op::SkipOne(okj, erj)
                 | Op::SkipMany(okj, erj, _)
@@ -825,15 +820,13 @@ impl<'a> PatternCompiler<'a> {
                 | Op::Upper(okj, erj)
                 | Op::Digit(okj, erj)
                 | Op::Alphanumeric(okj, erj)
-                | Op::Byte(okj, erj, _)
-                | Op::Bytes3(okj, erj, _) => {
-                    fix_jump(okj, index);
-                    fix_jump(erj, index);
+                | Op::Char(okj, erj, _)
+                | Op::String(okj, erj, _, _) => {
+                    fix_jump(okj, index, fix);
+                    fix_jump(erj, index, fix);
                 }
             }
         }
-
-        true
     }
 }
 
@@ -853,7 +846,7 @@ mod tests {
 
     #[test]
     fn assert_size() {
-        assert_eq!(8, std::mem::size_of::<Op>());
+        assert_eq!(16, std::mem::size_of::<Op>());
     }
 
     #[test]
@@ -941,6 +934,12 @@ mod tests {
         assert_eq!(MatchResult::Ok(1), p.matches("0"));
         assert_eq!(MatchResult::Ok(1), p.matches("9"));
         assert_eq!(MatchResult::Err, p.matches("!"));
+
+        let p = new_pattern("abcdefghij");
+        assert_eq!(MatchResult::Ok(10), p.matches("abcdefghij"));
+
+        let p = new_pattern("abcdefghijk");
+        assert_eq!(MatchResult::Ok(11), p.matches("abcdefghijk"));
     }
 
     #[test]
@@ -1026,6 +1025,12 @@ mod tests {
         assert_eq!(MatchResult::Err, p.matches("z"));
         assert_eq!(MatchResult::Err, p.matches("7a"));
         assert_eq!(MatchResult::Ok(3), p.matches("7ab"));
+        
+        let p = new_pattern("(abcdefghij)");
+        assert_eq!(MatchResult::Ok(10), p.matches("abcdefghij"));
+
+        let p = new_pattern("(abcdefghijk)");
+        assert_eq!(MatchResult::Ok(11), p.matches("abcdefghijk"));
     }
 
     #[test]
@@ -1206,6 +1211,19 @@ mod tests {
         assert_eq!(MatchResult::Ok(1), p.matches("cab"));
     }
 
+    #[test]
+    fn utf8() {
+        let p = new_pattern("[açé]");
+        assert_eq!(MatchResult::Ok(1), p.matches("a"));
+        assert_eq!(MatchResult::Ok('ç'.len_utf8()), p.matches("ç"));
+        assert_eq!(MatchResult::Ok('é'.len_utf8()), p.matches("é"));
+
+        let p = new_pattern(".");
+        assert_eq!(MatchResult::Ok(1), p.matches("a"));
+        assert_eq!(MatchResult::Ok('ç'.len_utf8()), p.matches("ç"));
+        assert_eq!(MatchResult::Ok('é'.len_utf8()), p.matches("é"));
+    }
+
     #[test]
     fn bad_pattern() {
         fn assert_err(expected: PatternError, value: Result<Pattern, PatternError>) {
@@ -1229,3 +1247,4 @@ mod tests {
         assert_err(PatternError::UnexpectedEndOfPattern, try_new_pattern("a|"));
     }
 }
+
