@@ -4,10 +4,7 @@ use std::{
         io::{AsRawFd, RawFd},
         net::{UnixListener, UnixStream},
     },
-    sync::{
-        atomic::{AtomicIsize, Ordering},
-        mpsc,
-    },
+    sync::atomic::{AtomicIsize, Ordering},
     time::Duration,
 };
 
@@ -20,8 +17,8 @@ use pepper::{
 
 mod unix_utils;
 use unix_utils::{
-    get_terminal_size, is_pipped, parse_terminal_keys, read, read_from_connection, run, Process,
-    RawMode,
+    errno, get_terminal_size, is_pipped, parse_terminal_keys, read, read_from_connection, run,
+    suspend_process, Process, RawMode,
 };
 
 const MAX_CLIENT_COUNT: usize = 20;
@@ -134,7 +131,7 @@ impl Kqueue {
             None => std::ptr::null(),
         };
 
-        let len = unsafe {
+        let mut len = unsafe {
             libc::kevent(
                 self.0,
                 [].as_ptr(),
@@ -145,7 +142,11 @@ impl Kqueue {
             )
         };
         if len == -1 {
-            panic!("could not wait for events");
+            if errno() == libc::EINTR {
+                len = 0;
+            } else {
+                panic!("could not wait for events");
+            }
         }
 
         events.0[..len as usize].iter().map(|e| {
@@ -192,7 +193,7 @@ fn run_server(args: Args, listener: UnixListener) -> Result<(), AnyError> {
         }
     }
 
-    let (request_sender, request_receiver) = mpsc::channel();
+    let (request_sender, request_receiver) = ServerApplication::platform_request_channel();
     let platform = Platform::new(flush_requests, request_sender);
     let event_sender = ServerApplication::run(args, platform);
 
@@ -249,29 +250,30 @@ fn run_server(args: Args, listener: UnixListener) -> Result<(), AnyError> {
                                 mut command,
                                 buf_len,
                             } => {
+                                let mut spawned = false;
                                 for (i, p) in processes.iter_mut().enumerate() {
                                     if p.is_some() {
                                         continue;
                                     }
 
                                     let handle = ProcessHandle(i);
-                                    match command.spawn() {
-                                        Ok(child) => {
-                                            let process = Process::new(child, tag, buf_len);
-                                            if let Some(fd) = process.try_as_raw_fd() {
-                                                kqueue
-                                                    .add(Event::Fd(fd), PROCESSES_START_INDEX + i);
-                                            }
-                                            *p = Some(process);
-                                            event_sender.send(
-                                                ApplicationEvent::ProcessSpawned { tag, handle },
-                                            )?;
+                                    if let Ok(child) = command.spawn() {
+                                        let process = Process::new(child, tag, buf_len);
+                                        if let Some(fd) = process.try_as_raw_fd() {
+                                            kqueue.add(Event::Fd(fd), PROCESSES_START_INDEX + i);
                                         }
-                                        Err(_) => event_sender
-                                            .send(ApplicationEvent::ProcessExit { tag })?,
+                                        *p = Some(process);
+                                        event_sender.send(ApplicationEvent::ProcessSpawned {
+                                            tag,
+                                            handle,
+                                        })?;
+                                        spawned = true;
                                     }
                                     break;
                                 }
+                                if !spawned {
+                                    event_sender.send(ApplicationEvent::ProcessExit { tag })?;
+                                }
                             }
                             PlatformRequest::WriteToProcess { handle, buf } => {
                                 let index = handle.0;
@@ -354,15 +356,10 @@ fn run_server(args: Args, listener: UnixListener) -> Result<(), AnyError> {
                         let tag = process.tag();
                         match process.read(&mut buf_pool) {
                             Ok(None) => (),
-                            Ok(Some(buf)) => {
-                                if buf.as_bytes().is_empty() {
-                                    event_sender.send(ApplicationEvent::ProcessExit { tag })?;
-                                } else {
-                                    event_sender
-                                        .send(ApplicationEvent::ProcessOutput { tag, buf })?;
-                                }
+                            Ok(Some(buf)) if !buf.as_bytes().is_empty() => {
+                                event_sender.send(ApplicationEvent::ProcessExit { tag })?;
                             }
-                            Err(()) => {
+                            _ => {
                                 if let Some(fd) = process.try_as_raw_fd() {
                                     kqueue.remove(Event::Fd(fd));
                                 }
@@ -400,7 +397,7 @@ fn run_client(args: Args, mut connection: UnixStream) {
         return;
     }
 
-    let raw_mode;
+    let mut raw_mode;
 
     let kqueue = Kqueue::new();
     kqueue.add(Event::Fd(connection.as_raw_fd()), 0);
@@ -414,12 +411,16 @@ fn run_client(args: Args, mut connection: UnixStream) {
         kqueue.add(Event::Resize, 2);
 
         let size = get_terminal_size();
-        let bytes = application.update(Some(size), &[], &[], &[]);
+        let (_, bytes) = application.update(Some(size), &[], &[], &[]);
         if connection.write(bytes).is_err() {
             return;
         }
     }
 
+    let backspace_code = match raw_mode {
+        Some(ref raw) => raw.backspace_code(),
+        None => 0,
+    };
     let mut keys = Vec::new();
     let mut buf = Vec::new();
 
@@ -451,7 +452,7 @@ fn run_client(args: Args, mut connection: UnixStream) {
                             if is_pipped {
                                 stdin_bytes = bytes;
                             } else {
-                                parse_terminal_keys(&bytes, &mut keys);
+                                parse_terminal_keys(bytes, backspace_code, &mut keys);
                             }
                         }
                     }
@@ -461,12 +462,16 @@ fn run_client(args: Args, mut connection: UnixStream) {
                 Err(()) => break 'main_loop,
             }
 
-            let bytes = application.update(resize, &keys, stdin_bytes, server_bytes);
+            let (suspend, bytes) = application.update(resize, &keys, stdin_bytes, server_bytes);
             if connection.write(bytes).is_err() {
                 break;
             }
+            if suspend {
+                suspend_process(&mut application, &mut raw_mode);
+            }
         }
     }
 
     drop(raw_mode);
 }
+
