@@ -3,10 +3,7 @@ use std::{
     os::windows::{ffi::OsStrExt, io::IntoRawHandle},
     process::Child,
     ptr::NonNull,
-    sync::{
-        atomic::{AtomicPtr, Ordering},
-        mpsc,
-    },
+    sync::atomic::{AtomicPtr, Ordering},
     time::Duration,
 };
 
@@ -886,7 +883,7 @@ fn run_server(args: Args, pipe_path: &[u16]) -> Result<(), AnyError> {
     let new_request_event = Event::automatic();
     NEW_REQUEST_EVENT_HANDLE.store(new_request_event.0 as _, Ordering::Relaxed);
 
-    let (request_sender, request_receiver) = mpsc::channel();
+    let (request_sender, request_receiver) = ServerApplication::platform_request_channel();
     let mut platform = Platform::new(
         || set_event(NEW_REQUEST_EVENT_HANDLE.load(Ordering::Relaxed) as _),
         request_sender,
@@ -941,26 +938,24 @@ fn run_server(args: Args, pipe_path: &[u16]) -> Result<(), AnyError> {
                             mut command,
                             buf_len,
                         } => {
+                            let mut spawned = false;
                             for (i, p) in processes.iter_mut().enumerate() {
                                 if p.is_some() {
                                     continue;
                                 }
 
                                 let handle = ProcessHandle(i);
-                                match command.spawn() {
-                                    Ok(child) => {
-                                        *p = Some(AsyncProcess::new(child, tag, buf_len));
-                                        event_sender.send(ApplicationEvent::ProcessSpawned {
-                                            tag,
-                                            handle,
-                                        })?;
-                                    }
-                                    Err(_) => {
-                                        event_sender.send(ApplicationEvent::ProcessExit { tag })?
-                                    }
+                                if let Ok(child) = command.spawn() {
+                                    *p = Some(AsyncProcess::new(child, tag, buf_len));
+                                    event_sender
+                                        .send(ApplicationEvent::ProcessSpawned { tag, handle })?;
+                                    spawned = true;
                                 }
                                 break;
                             }
+                            if !spawned {
+                                event_sender.send(ApplicationEvent::ProcessExit { tag })?;
+                            }
                         }
                         PlatformRequest::WriteToProcess { handle, buf } => {
                             if let Some(ref mut process) = processes[handle.0] {
@@ -1025,15 +1020,10 @@ fn run_server(args: Args, pipe_path: &[u16]) -> Result<(), AnyError> {
                         let tag = process.tag;
                         match pipe.read_async(&mut buf_pool) {
                             Ok(None) => (),
-                            Ok(Some(buf)) => {
-                                if buf.as_bytes().is_empty() {
-                                    event_sender.send(ApplicationEvent::ProcessExit { tag })?;
-                                } else {
-                                    event_sender
-                                        .send(ApplicationEvent::ProcessOutput { tag, buf })?;
-                                }
+                            Ok(Some(buf)) if !buf.as_bytes().is_empty() => {
+                                event_sender.send(ApplicationEvent::ProcessOutput { tag, buf })?;
                             }
-                            Err(()) => {
+                            _ => {
                                 process.stdout = None;
                                 process.kill();
                                 processes[i] = None;
@@ -1192,7 +1182,7 @@ fn run_client(args: Args, pipe_path: &[u16], input_handle: Handle, output_handle
                 console_output_mode = Some(output_mode);
 
                 let size = get_console_size(output_handle);
-                let bytes = application.update(Some(size), &[], &[], &[]);
+                let (_, bytes) = application.update(Some(size), &[], &[], &[]);
                 if !connection.write(bytes) {
                     return;
                 }
@@ -1243,7 +1233,7 @@ fn run_client(args: Args, pipe_path: &[u16], input_handle: Handle, output_handle
             _ => unreachable!(),
         }
 
-        let bytes = application.update(resize, &keys, stdin_bytes, server_bytes);
+        let (_, bytes) = application.update(resize, &keys, stdin_bytes, server_bytes);
         if !connection.write(bytes) {
             break;
         }
@@ -1329,3 +1319,4 @@ fn parse_console_events(
         }
     }
 }
+
